import 'dart:io';
import "dart:mirrors" hide SourceLocation;
import 'package:analysis_utils/src/expression_handler.dart';
import 'package:analysis_utils/src/resolver.dart';
import "package:analyzer/analyzer.dart";
import 'package:analyzer/dart/ast/token.dart';
import 'package:source_span/source_span.dart';

/// An analyzed Dart file/code
///
/// All what must be parsed with the analysis package should be done with this
class SourceAnalysis {
  static final Map<String, SourceAnalysis> _analysisCache = {};

  /// The path in a String manner
  String path;

  /// The result of executing the `parseXXX()`'s Analysis package methods
  CompilationUnit compilationUnit;

  String _code;

  List<ClassAnalysis> _classes;
  List<TopLevelVarAnalysis> _topLevelVariables;

  /// The code of the file
  String get code {
    if (_code == null) {
      File file = new File(path);
      _code = file.readAsStringSync();
    }
    return _code;
  }

  /// Entrypoint for static code. You can use it to analyze the declarations (TopLevelVariables
  /// and Classes) of the dart code provided in `contents`
  factory SourceAnalysis.forContents(String contents) {
    if (_analysisCache.containsKey(contents))
      return _analysisCache[contents];
    else {
      final SourceAnalysis ret = new SourceAnalysis._forContents(contents);
      _analysisCache[contents] = ret;
      return ret;
    }
  }

  factory SourceAnalysis.forFilePath(String path) {
    if (_analysisCache.containsKey(path))
      return _analysisCache[path];
    else {
      final SourceAnalysis ret = new SourceAnalysis._(path);
      _analysisCache[path] = ret;
      return ret;
    }
  }

  factory SourceAnalysis.forMirror(DeclarationMirror mirror) {
    Uri uri = mirror.location?.sourceUri;
    if (uri == null)
      throw new UnsupportedError(
          "The mirror '$mirror' doesn't support the sourceUri property. Try providing a SourceAnalysis instance from another mirror that does.");
    String path = resolvePath(uri);
    return new SourceAnalysis.forFilePath(path);
  }

  SourceAnalysis._(this.path) {
    if (path.endsWith(".dart")) {
      this.compilationUnit = parseDartFile(path);
    } else
      throw new Exception("You can only parse .dart files");
  }

  SourceAnalysis._forContents(String contents) {
    this._code = contents;
    this.compilationUnit = parseCompilationUnit(contents);
  }

  List<ClassAnalysis> get classes {
    if (_classes != null)
      return _classes;
    else {
      _classes = [];
      compilationUnit.declarations
          .whereType<ClassDeclaration>()
          .forEach((ClassDeclaration classDeclaration) {
        ClassAnalysis analysis =
            new ClassAnalysis.fromAnalysis(classDeclaration, source: this);
        _classes.add(analysis);
      });
      return _classes;
    }
  }

  List<TopLevelVarAnalysis> get topLevelVariables {
    if (_topLevelVariables != null)
      return _topLevelVariables;
    else {
      _topLevelVariables = [];
      compilationUnit.declarations
          .whereType<TopLevelVariableDeclaration>()
          .forEach((TopLevelVariableDeclaration tlv) {
        tlv.variables.variables.forEach((VariableDeclaration v) {
          TopLevelVarAnalysis analysis =
              new TopLevelVarAnalysis.fromAnalysis(tlv, v, this);
          _topLevelVariables.add(analysis);
        });
      });
      return _topLevelVariables;
    }
  }
}

/// base class used for traversing Analysis nodes in oprder to find the needed
/// ones (allocated in `finded`)
abstract class Finder<D> extends SimpleAstVisitor {
  final String name;
  D finded;
  Finder(this.name);
}

class ParameterFinder extends Finder<FormalParameterList> {
  ParameterFinder(String name) : super(name) {}

  // @override
  // visitImportDirective(ImportDirective node) => null;
  // @override
  // visitComment(node) => null;
  // @override
  // visitTypeName(node) => print("TypeName: $node");
  // @override
  // visitSimpleIdentifier(node) => print("SimpleIdentifier: $node");
  // @override
  visitFormalParameterList(FormalParameterList node) {
    if (finded != null)
      throw new Exception("There can't be 2 parametersList, WTF?");
    this.finded = node;
  }
  //
  // @override
  // visitExpressionFunctionBody(node) => print("FunctionBody: $node");
}

class ClassFinder extends Finder<ClassDeclaration> {
  ClassFinder(String name) : super(name);

  @override
  visitClassDeclaration(ClassDeclaration node) {
    if (node.name.toString() == name) {
      this.finded = node;
    }
    return super.visitClassDeclaration(node);
  }
}

class FieldFinder extends Finder<FieldDeclaration> {
  FieldFinder(String name) : super(name);
  @override
  visitFieldDeclaration(FieldDeclaration node) {
    if (node.fields.variables
        .any((VariableDeclaration v) => v.name.toString() == name)) {
      this.finded = node;
    }
    return super.visitFieldDeclaration(node);
  }
}

class ConstructorFinder extends Finder<ConstructorDeclaration> {
  ConstructorFinder(String name)
      : super(name.startsWith(".") ? name.substring(1) : name);
  visitConstructorDeclaration(ConstructorDeclaration node) {
    if ((node.name?.toString() ?? "") == name) {
      this.finded = node;
    }
    return super.visitConstructorDeclaration(node);
  }
}

class MethodFinder extends Finder<MethodDeclaration> {
  MethodFinder(String name) : super(name);
  @override
  visitMethodDeclaration(MethodDeclaration node) {
    if (node.name.toString() == name) this.finded = node;
    return super.visitMethodDeclaration(node);
  }
}

class TopLevelVariableFinder extends Finder<TopLevelVariableDeclaration> {
  TopLevelVariableFinder(String name) : super(name);
  @override
  visitTopLevelVariableDeclaration(TopLevelVariableDeclaration node) {
    if (node.variables.variables
        .any((VariableDeclaration d) => d.name.name == this.name)) {
      this.finded = node;
      return super.visitTopLevelVariableDeclaration(node);
    }
  }
}

abstract class EntityAnalysis<A extends AstNode, F extends Finder,
    M extends DeclarationMirror> {
  String name;
  String docs;
  SourceSpan location;
  List<MetadataAnalysis> metadata;
  SourceAnalysis source;
  M mirror;
  A analyzerDeclaration;
  F _entityFinder;

  /// the analyzer node tha contains this entity. For example a FieldDeclaration
  /// should be contained by a ClassDeclaration
  Declaration _analyzerContainer;

  List<MetadataAnalysis> _computeMetadata() {
    NodeList<Annotation> annotations =
        (analyzerDeclaration as AnnotatedNode).metadata;
    List<MetadataAnalysis> ret = [];
    for (int i = 0; i < annotations.length; i++) {
      Annotation a = annotations[i];
      if (mirror == null) {
        ret.add(new MetadataAnalysis.fromAnalysis(a));
      } else {
        ret.add(new MetadataAnalysis(a, mirror.metadata[i]));
      }
    }
    return ret;
  }

  EntityAnalysis.fromAnalysis(this.analyzerDeclaration, this.source) {
    _computeAnalysis();
    this.location = _computeLocation();
  }

  EntityAnalysis(this.mirror, {Declaration analyzerContainer, this.source}) {
    this._analyzerContainer = analyzerContainer;
    _computeName();
    this.source ??= new SourceAnalysis.forMirror(mirror);
    _findDeclaration();
    _computeAnalysis();
    this.location = _computeLocation();
  }

  void _computeName() {
    this.name = MirrorSystem.getName(mirror.simpleName);
  }

  String toString() => name;

  _findDeclaration() {
    ClassMirror finderMirror = reflectClass(F);
    this._entityFinder =
        finderMirror.newInstance(new Symbol(""), [name]).reflectee;
    if (_analyzerContainer != null)
      _analyzerContainer.visitChildren(_entityFinder);
    else
      source.compilationUnit.visitChildren(_entityFinder);
    this.analyzerDeclaration = _entityFinder.finded as A;
  }

  _computeAnalysis() {
    if (analyzerDeclaration != null && analyzerDeclaration is AnnotatedNode) {
      this.docs = _computeDocs(analyzerDeclaration as AnnotatedNode);
      this.metadata = _computeMetadata();
    }
  }

  SourceSpan _computeLocation() {
    var node;
    try {
      node = (analyzerDeclaration as AnnotatedNode);
    } catch (e) {
      node = analyzerDeclaration;
    }
    Token startToken = ((node is AnnotatedNode)
                ? node?.documentationComment?.beginToken
                : null) ??
            node?.beginToken,
        endToken = node?.endToken;
    if (startToken == null || endToken == null) return null;
    SourceLocation start =
        new SourceLocation(startToken.offset, sourceUrl: source.path);
    SourceLocation end =
        new SourceLocation(endToken.offset + 1, sourceUrl: source.path);
    String text = source.code.substring(start.offset, end.offset);
    return new SourceSpan(start, end, text);
  }

  String _computeDocs(AnnotatedNode declaration) {
    List<Token> tokens = declaration?.documentationComment?.tokens;
    if (tokens == null || tokens.length == 0) return "";
    if (tokens.length == 1) {
      //Maybe a /** docs */ documentation type?
      String doc = tokens.first.toString();
      if (doc.startsWith("/**"))
        return _computeComplexDocsComment(doc);
      else if (doc.startsWith("///"))
        return _computeNormalDocsComment(tokens);
      else
        throw new UnsupportedError(
            "Don't kwno what to do here. Please make me an issue in analysis_utils package showing the documentation of '$name' entity");
    } else {
      // Must be the classical /// documentation
      return _computeNormalDocsComment(tokens);
    }
  }

  String _computeComplexDocsComment(String docLine) {
    RegExp formatter = new RegExp("\n[ *]*");
    return docLine
        .substring(3, docLine.length - 3) //removes the /** and */
        .replaceAll(formatter, " ")
        .trim();
  }

  String _computeNormalDocsComment(List<Token> tokens) {
    return tokens
        .map((Token t) => t.toString().substring(3).trim())
        .join(" ")
        .trim();
  }
}

class MetadataAnalysis {
  Type type;
  String name;
  Annotation node;
  dynamic instance;
  InstanceMirror mirror;
  ArgumentsResolution arguments;

  MetadataAnalysis(this.node, this.mirror) {
    this.instance = mirror.reflectee;
    this.type = mirror.type.reflectedType;
    _processGenerics();
  }

  MetadataAnalysis.fromAnalysis(this.node) {
    _processGenerics();
  }

  void _processGenerics() {
    this.name = node.name.name;
    this.arguments = new ArgumentsResolution.fromArgumentList(node.arguments);
  }
}

class TopLevelVarAnalysis extends EntityAnalysis<TopLevelVariableDeclaration,
    TopLevelVariableFinder, VariableMirror> {
  var defaultValue;
  String typeString;
  VariableDeclaration variableDeclaration;
  TypeAnnotation typeAnalysis;

  TopLevelVarAnalysis.fromAnalysis(TopLevelVariableDeclaration topLevelVariable,
      this.variableDeclaration, SourceAnalysis source)
      : super.fromAnalysis(topLevelVariable, source) {
    this.name = variableDeclaration.name.name;
    this.defaultValue =
        variableDeclaration.initializer?.accept(new ConstantEvaluator());
    this.typeAnalysis = topLevelVariable.variables.type;
    if (typeAnalysis != null) {
      this.typeString =
          typeAnalysis.type?.name ?? typeAnalysis.toString() ?? "var";
      if (typeAnalysis is NamedType &&
          (typeAnalysis as NamedType).typeArguments != null &&
          !typeString.contains("<")) {
        this.typeString +=
            "<${(typeAnalysis as NamedType).typeArguments.arguments.join(",")}>";
      }
    } else {
      this.typeString = "var";
    }
  }
}

class ClassAnalysis
    extends EntityAnalysis<ClassDeclaration, ClassFinder, ClassMirror> {
  ClassAnalysis superclass;
  Map<String, FieldAnalysis> fields = {};
  Map<String, MethodAnalysis> methods = {};
  Map<String, ConstructorAnalysis> constructors = {};
  InstanceMirror instanceMirror;

  factory ClassAnalysis.fromInstance(Object instance) {
    InstanceMirror instanceMirror = reflect(instance);
    return new ClassAnalysis.fromMirror(instanceMirror.type, instanceMirror);
  }

  factory ClassAnalysis.fromType(Type type) {
    ClassMirror mirror = reflectClass(type);
    return new ClassAnalysis.fromMirror(mirror);
  }

  ClassAnalysis.fromMirror(ObjectMirror mirror, [this.instanceMirror])
      : super(mirror as ClassMirror) {
    ClassMirror classMirror = (mirror as ClassMirror);
    this.superclass = classMirror.superclass == null ||
            classMirror.superclass.reflectedType == Object
        ? null
        : new ClassAnalysis.fromMirror(classMirror.superclass);
    Map<String, FieldAnalysis> fields = {};
    Map<String, MethodAnalysis> methods = {};
    classMirror.declarations.forEach((Symbol name, DeclarationMirror d) {
      // if (d is VariableMirror) this.fields.add(new FieldAnalysis.fromMirror(d));
      // if (d is MethodMirror) this.methods.add(new MethodAnalysis.fromMirror(d));
      if (d is VariableMirror) {
        FieldAnalysis analysis = new FieldAnalysis(this, d);
        fields[analysis.name] = analysis;
      }
      if (d is MethodMirror) {
        if (d.isConstructor) {
          var analysis = new ConstructorAnalysis(this, d);
          this.constructors[analysis.name] = analysis;
        } else {
          var analysis = new MethodAnalysis(this, d);
          methods[analysis.name] = analysis;
        }
      }
    });
    if (superclass != null) {
      this.fields = new Map.fromEntries(superclass.fields.entries);
      this.methods = new Map.fromEntries(superclass.methods.entries);
    }
    this.fields.addAll(fields);
    this.methods.addAll(methods);
  }

  ClassAnalysis.fromAnalysis(ClassDeclaration classDeclaration,
      {SourceAnalysis source})
      : super.fromAnalysis(classDeclaration, source) {
    this.name = classDeclaration.name.name;
    classDeclaration.members.forEach((ClassMember member) {
      if (member is FieldDeclaration) {
        member.fields.variables.forEach((VariableDeclaration declaration) {
          FieldAnalysis field =
              new FieldAnalysis.fromAnalysis(this, member, declaration);
          this.fields[field.name] = field;
        });
      } else if (member is ConstructorDeclaration) {
        ConstructorAnalysis analysis =
            new ConstructorAnalysis.fromAnalysis(this, member);
        this.constructors[analysis.name ?? ""] = analysis;
      } else if (member is MethodDeclaration) {
        MethodAnalysis analysis = new MethodAnalysis.fromAnalysis(this, member);
        this.methods[analysis.name] = analysis;
      }
    });
  }

  ///Filter the methods that aren't getters or setters only
  Map<String, MethodAnalysis> get regularMethods => new Map.fromEntries(
      methods.entries.where((m) => m.value.mirror.isRegularMethod));

  ///Filter the getter methods only
  Map<String, MethodAnalysis> get getters => new Map.fromEntries(
      methods.entries.where((m) => m.value.mirror.isGetter));

  ///Filter the setter methods only
  Map<String, MethodAnalysis> get setters => new Map.fromEntries(
      methods.entries.where((m) => m.value.mirror.isSetter));

  ///Filter the static methods only
  Map<String, MethodAnalysis> get staticMethods => new Map.fromEntries(
      methods.entries.where((m) => m.value.mirror.isStatic));

  bool get isEnum => this.mirror.isEnum;

  Map<String, dynamic> toMap() {
    Map ret = this.fields.map<String, dynamic>(
        (k, v) => new MapEntry(k, v.getInstanceMirror.reflectee));
    return ret;
  }
}

class ClassMemberAnalysis<D extends ClassMember, F extends Finder,
    M extends DeclarationMirror> extends EntityAnalysis<D, F, M> {
  ClassAnalysis container;
  ClassMemberAnalysis(
      DeclarationMirror mirror, this.container, SourceAnalysis sourceAnalysis)
      : super(mirror,
            analyzerContainer: container.analyzerDeclaration,
            source: sourceAnalysis);

  ClassMemberAnalysis.fromAnalysis(
      D declaration, this.container, SourceAnalysis sourceAnalysis)
      : super.fromAnalysis(declaration, sourceAnalysis);
}

/// Analyzes the variables declarations inside classes scope (fields)
/// The [FieldDeclaration]s have the vissicitude of being able to be declared
/// with several variables in a single field (for example: 'Type var1 = val1,
/// var2' will have 2 [VariableDeclaration]s declared in the same field.
/// While the [VariableMirror] doesn't do that distinction at all. In my
/// implementation of this [FieldAnalysis] class, the `analyzerDeclaration` will
/// hold the field, while `variableDeclaration` the variable
class FieldAnalysis
    extends ClassMemberAnalysis<FieldDeclaration, FieldFinder, VariableMirror> {
  TypeMirror typeMirror;
  String typeString;
  NamedType typeAnalysis;

  VariableDeclaration variableDeclaration;

  bool _hasInstanceMirror;
  ClassMirror _classMirror;

  ClassMirror get classMirror {
    _classMirror ??= reflectClass(typeMirror.reflectedType);
    return _classMirror;
  }

  /// If there's an instance mirror, get that instance's this field value and
  /// return a mirror on that result
  InstanceMirror get getInstanceMirror {
    if (_hasInstanceMirror) {
      return container.instanceMirror.getField(new Symbol(this.name));
    } else {
      throw new StateError(
          "Requesting instanceMirror of a field whose ClassAnalysis's "
          "parent container was declared without .fromInstance constructor");
    }
  }

  /// If there's an instance mirror, set that instance's this field value and
  /// return a mirror on the result
  InstanceMirror setInstanceMirror(value) {
    if (_hasInstanceMirror) {
      return container.instanceMirror.setField(new Symbol(this.name), value);
    } else {
      throw new StateError(
          "Requesting instanceMirror of a field whose ClassAnalysis's "
          "parent container was declared without .fromInstance constructor");
    }
  }

  bool get isFinal => this.variableDeclaration.isFinal;
  bool get isConst => this.variableDeclaration.isConst;
  bool get isPrivate => this.mirror?.isPrivate ?? this.name.startsWith("_");

  FieldAnalysis.fromAnalysis(ClassAnalysis container,
      FieldDeclaration fieldDeclaration, this.variableDeclaration)
      : super.fromAnalysis(fieldDeclaration, container, container.source) {
    this.name = variableDeclaration.name.name;
    if (fieldDeclaration.fields.type is NamedType) {
      this.typeAnalysis = fieldDeclaration.fields.type;
      this.typeString = typeAnalysis.name.name;
      if (typeAnalysis.typeArguments != null) {
        this.typeString +=
            "<${typeAnalysis.typeArguments.arguments.join(",")}>";
      }
    }
  }

  FieldAnalysis(ClassAnalysis container, VariableMirror declaration)
      : super(declaration, container, container.source) {
    this.variableDeclaration = analyzerDeclaration?.fields?.variables
        ?.singleWhere(
            (VariableDeclaration d) => d.name.toString() == this.name);
    this.typeMirror = declaration.type;
    this._hasInstanceMirror = container.instanceMirror != null;
  }

  List<String> get typeGenerics {
    var t = analyzerDeclaration.fields.type as NamedType;
    return t.typeArguments?.arguments?.map((t) => t.toString())?.toList() ?? [];
  }

  get defaultValue =>
      variableDeclaration.initializer?.accept(new ConstantEvaluator());
}

class ConstructorAnalysis extends ClassMemberAnalysis<ConstructorDeclaration,
    ConstructorFinder, MethodMirror> with ParametersInterface {
  ConstructorDeclaration analyzerDeclaration;
  ConstructorAnalysis(ClassAnalysis container, MethodMirror declaration)
      : super(declaration, container, container.source) {
    this.parameters = new ParametersAnalysis(this, declaration);
  }

  ConstructorAnalysis.fromAnalysis(
      ClassAnalysis classAnalysis, ConstructorDeclaration declaration)
      : super.fromAnalysis(declaration, classAnalysis, classAnalysis.source) {
    this.parameters =
        new ParametersAnalysis.fromAnalysis(this, declaration.parameters);
  }

  bool get isConst => analyzerDeclaration.constKeyword != null;
  bool get isExternal => analyzerDeclaration.externalKeyword != null;
  bool get isFactory => analyzerDeclaration.factoryKeyword != null;
  bool get isPrivate => mirror.isPrivate;

  _computeName() {
    String str = MirrorSystem.getName(this.mirror.simpleName);
    int index = str.indexOf(r".");
    this.name = index == -1 ? "" : str.substring(index + 1);
  }
}

class MethodAnalysis
    extends ClassMemberAnalysis<MethodDeclaration, MethodFinder, MethodMirror>
    with ParametersInterface {
  MethodAnalysis(ClassAnalysis container, MethodMirror declaration)
      : super(declaration, container, container.source) {
    this.parameters = new ParametersAnalysis(this, declaration);
    this.name = MirrorSystem.getName(declaration.simpleName);
  }

  MethodAnalysis.fromAnalysis(
      ClassAnalysis classAnalysis, MethodDeclaration member)
      : super.fromAnalysis(member, classAnalysis, classAnalysis.source) {
    this.parameters =
        new ParametersAnalysis.fromAnalysis(this, member.parameters);
    this.name = member.name.name;
  }

  bool get isPrivate => mirror.isPrivate;

  InstanceMirror invocationMirror(
      List positionalArguments, Map<Symbol, dynamic> namedArguments) {
    if (container.instanceMirror == null) {
      throw new StateError(
          "Requesting instanceMirror of a method whose ClassAnalysis's "
          "parent container was declared without .fromInstance constructor");
    } else {
      return container.instanceMirror
          .invoke(new Symbol(name), positionalArguments, namedArguments);
    }
  }
}

class ParametersInterface {
  ParametersAnalysis parameters;

  /// Syntactic sugar for the set of always optional parameters that need a name
  /// to be invoked in arguments
  Set<Parameter> get namedParameters => parameters.named;

  /// Syntactic sugar for the always required, ordered, positional parameters
  List<Parameter> get requiredParameters => parameters.ordinary;

  /// Syntactic sugar for the optionals (both named as positional) parameters
  List<Parameter> get optionalParameters => parameters.optionals;

  /// Same as `requiredParameters`, the normal old school lifelonging parameters
  List<Parameter> get ordinaryParameters => parameters.ordinary;

  /// Syntactic sugar for the List of parameters that should (or could) be
  /// invoked (both the types of required and the optionals)
  List<Parameter> get positionalParameters => parameters.positionals;

  /// Syntactic sugar for the List of positional optional parameters (those
  /// defined between square brackets)
  List<Parameter> get positionalOptionalParameters =>
      parameters.positionalOptionals;
}

class Parameter {
  var defaultValue;
  Type type;
  String typeString;
  String name;
  ParameterMirror mirror;
  FormalParameter node;

  Parameter(FormalParameterList container, this.mirror) {
    this.type = mirror.type.reflectedType;
    this.name = MirrorSystem.getName(mirror.simpleName);
    this.defaultValue =
        mirror.hasDefaultValue ? mirror.defaultValue.reflectee : null;
    try {
      this.node = container.parameters.singleWhere(
          (FormalParameter parameter) =>
              parameter.identifier.toString() == this.name);
    } catch (e) {
      throw new Exception(
          "Weird error... there seems to not exist a parameter node for a mirror of itself");
    }
  }

  Parameter.fromAnalysis(this.node) {
    this.name = node.identifier.toString();
    if (node is DefaultFormalParameter) {
      this.defaultValue = (node as DefaultFormalParameter)
          ?.defaultValue
          ?.accept(new ConstantEvaluator());
      if (defaultValue == NOT_A_CONSTANT) {
        this.defaultValue = null;
      }
      this.node = (node as DefaultFormalParameter).parameter;
    }
    if (node is SimpleFormalParameter) {
      this.typeString = (node as SimpleFormalParameter)?.type?.toString();
    } else if (node is FieldFormalParameter) {
      this.typeString = (node as FieldFormalParameter)?.type?.type?.displayName;
    } else {
      throw new UnimplementedError(
          "No programé qué hacer con ${node.runtimeType}");
    }
  }

  bool get isOptional => node.isOptional;
  bool get isNamed => node.isNamed;
  bool get isOrdinary => node.isRequired;
  bool get isPositionalOptional => node.isOptionalPositional;
  bool get isThisInitializer => node is FieldFormalParameter;
  // bool get isThisInitializer => node.beginToken.toString() == "this";
}

class ParametersAnalysis
    extends EntityAnalysis<FormalParameterList, ParameterFinder, MethodMirror> {
  ClassMemberAnalysis container;
  List<Parameter> _parameters = [];

  ParametersAnalysis(this.container, MethodMirror mirror)
      : super(mirror,
            source: container.source,
            analyzerContainer: container.analyzerDeclaration) {
    if (analyzerDeclaration != null) {
      mirror.parameters.forEach((ParameterMirror m) {
        this._parameters.add(new Parameter(analyzerDeclaration, m));
      });
    }
  }

  ParametersAnalysis.fromAnalysis(
      this.container, FormalParameterList parameters)
      : super.fromAnalysis(parameters, container.source) {
    this.analyzerDeclaration.parameters.forEach((FormalParameter param) {
      this._parameters.add(new Parameter.fromAnalysis(param));
    });
  }

  /// All the parameters, simple as they are
  List<Parameter> get all => _parameters;

  /// Always required, ordered, positional parameters
  List<Parameter> get required =>
      _parameters.where((p) => p.isOrdinary).toList();

  /// Optional parameters (both positional and named)
  List<Parameter> get optionals =>
      _parameters.where((p) => p.isOptional).toList();

  /// Set of always optional parameters that need a name to be invoked in
  /// arguments
  Set<Parameter> get named => _parameters.where((p) => p.isNamed).toSet();

  /// The normal old school lifelonging parameters (required positionals)
  List<Parameter> get ordinary =>
      _parameters.where((p) => p.isOrdinary).toList();

  /// List of parameters that should (or could) be invoked in a stablished order
  /// (both the required and the optionals types)
  List<Parameter> get positionals =>
      _parameters.where((p) => p.isOrdinary || p.isPositionalOptional).toList();

  /// The positional optional parameters (those defined betwenn square brackets)
  List<Parameter> get positionalOptionals =>
      _parameters.where((p) => p.isPositionalOptional).toList();

  get length => _parameters.length;

  operator [](String name) =>
      _parameters.singleWhere((Parameter p) => p.name == name, orElse: null);
}
